#include "pdf.h"
#include "dyres_interface.h"
#include "interface.h"
#include "scales.h"
#include "settings.h"
#include "coupling.h"
#include "alphas.h"
#include "isnan.h"
#include "vjint.h"
#include "mcfm_interface.h"
#include "resconst.h"
#include "constants.h"
#include "parton.h"

#include <LHAPDF/LHAPDF.h>
#include <LHAPDF/LHAGlue.h>
#include <math.h>

using namespace constants;

bool pdf::secondbeam = false;
LHAPDF::PDF* pdf::lhapdf_beam1 = 0;
LHAPDF::PDF* pdf::lhapdf_beam2 = 0;

int pdf::order;
double pdf::xmin;
double pdf::qmin = 0.;
double pdf::mc;
double pdf::mb;
double pdf::mt;
double pdf::g1;
double pdf::g2;
double pdf::e;
double pdf::g0;
double pdf::q;
double pdf::blim;
double pdf::lambda;
bool pdf::photon = false;

void (*pdf::xfxq)(const double &x, const double &Q, double *fPDF) = 0;
void (*pdf::xfxqphot)(const double &x, const double &Q, double &fotfx) = 0;
double (*pdf::extalphas)(const double &Q) = 0;

void pdf::lhaxfxq_beam1(const double &x, const double &Q, double *r)
{
  vector <double> fPDF; fPDF.resize(13);
  lhapdf_beam1->xfxQ(x,Q,fPDF);
  copy(fPDF.begin(),fPDF.end(),r);
}

void pdf::lhaxfxq_beam2(const double &x, const double &Q, double *r)
{
  vector <double> fPDF; fPDF.resize(13);
  lhapdf_beam2->xfxQ(x,Q,fPDF);
  copy(fPDF.begin(),fPDF.end(),r);
}

void pdf::lhaxfxqphot_beam1(const double &x, const double &Q, double &fotfx)
{
  fotfx = pdf::lhapdf_beam1->xfxQ(22,x,Q)/x;
  //fotfx = lhapdf_beam1->xfxQ(22,x,Q)/x;
}

void pdf::lhaxfxqphot_beam2(const double &x, const double &Q, double &fotfx)
{
  fotfx = lhapdf_beam2->xfxQ(22,x,Q)/x;
}

void pdf::lhaxfxqphot_empty(const double &x, const double &Q, double &fotfx)
{
  fotfx = 0.;
}

double pdf::lhaalphas(const double &Q)
{
  //Old LHAPDF interface
  //return LHAPDF::alphasPDF(Q);

  //New LHAPDF interface
  return lhapdf_beam1->alphasQ(Q);
}

double pdf::rgktalphas(double q)
{
  //Interface to the Runge-Kutta alphas solution, with order matched to opts.order
  //This is always the MSbar alphas

  //Take alphas(mZ) from the external (lhapdf) alphas
  double q0 = coupling::zmass;
  double as0 = extalphas(q0);

  if (opts.order == 0)
    return alphas::lo(q,q0,as0);

  return real(alphas::exact(q,q0,as0,opts.order+1));
  
  //Evolve from mZ to mu with Runge-Kutta (MSbar evolution)
  alphas::rgkt(q,q0,as0);
  switch (opts.order)
    {
    case 0: return real(alphas::asLO);     break;
    case 1: return real(alphas::asNLO);    break;
    case 2: return real(alphas::asNNLO);   break;
    case 3: return real(alphas::asNNNLO);  break;
    case 4: return real(alphas::asNNNNLO); break;
    }
}

void pdf::init()
{
  if (!opts.externalpdf)
    {
      //printf(" ==Initialize PDF set from LHAPDF==\n\n");
      //printf("\n");
      LHAPDF::Info& cfg = LHAPDF::getConfig();

      cfg.set_entry("Verbosity"             , 0              );
      cfg.set_entry("Interpolator"          , "logcubic"     );
      cfg.set_entry("Extrapolator"          , "continuation" );
      cfg.set_entry("ForcePositive"         , 0              );
      cfg.set_entry("AlphaS_Type"           , "analytic"     );
      cfg.set_entry("MZ"                    , 91.1876        );
      cfg.set_entry("MUp"                   , 0.002          );
      cfg.set_entry("MDown"                 , 0.005          );
      cfg.set_entry("MStrange"              , 0.10           );
      cfg.set_entry("MCharm"                , 1.29           );
      cfg.set_entry("MBottom"               , 4.19           );
      cfg.set_entry("MTop"                  , 172.9          );
      cfg.set_entry("Pythia6LambdaV5Compat" , true           );

      //Old interface
      LHAPDF::initPDFSet(opts.LHAPDFset); //LHAPDF::initPDFSetByName(opts.LHAPDFset);
      LHAPDF::initPDF(opts.LHAPDFmember);

      //New interface
      lhapdf_beam1 = LHAPDF::mkPDF(opts.LHAPDFset, opts.LHAPDFmember);

      if (opts.LHAPDFset_beam2 != "none")
	{
	  lhapdf_beam2 = LHAPDF::mkPDF(opts.LHAPDFset_beam2, opts.LHAPDFmember_beam2);
	  secondbeam = true;
	}
      else
	lhapdf_beam2 = LHAPDF::mkPDF(opts.LHAPDFset, opts.LHAPDFmember);
      
      xfxq = lhaxfxq_beam1;
      
      //if (LHAPDF::hasPhoton())
      if (lhapdf_beam1->hasFlavor(22))
	{
	  photon = true;
	  xfxqphot = lhaxfxqphot_beam1;
	}
      else
	{
	  photon = false;
	  xfxqphot = lhaxfxqphot_empty;
	}
      
      //if (opts.lhaalphas)
      extalphas = lhaalphas;
      //else
      //alphas = rgktalphas
//      if (opts.PDFerrors && LHAPDF::numberPDF() > 1)
//	{
//	  opts.totpdf = LHAPDF::numberPDF()+1;
//	  pdferropts_.pdferr_ = true;
//	  pdferropts_.totpdf_ = LHAPDF::numberPDF()+1;
//	}
//      else
//	{
	  opts.totpdf = 1;
	  pdferropts_.pdferr_ = false;
	  pdferropts_.totpdf_ = 1;
	  //	}

      order = LHAPDF::getOrderPDF(); //order of evolution

      //set order_evol not smaller than the order of the PDF:
      if (opts.orderlha)
	//opts.order_evol = max(opts.order_evol,pdf::order+1);
	opts.order_evol = pdf::order+1;
      
      LHAPDF::PDFInfo info(opts.LHAPDFset, opts.LHAPDFmember);
      qmin = info.get_entry_as<double>("QMin", -1);
      xmin = info.get_entry_as<double>("XMin", -1);

      //Set blim_pdf not larger than b0/Qmin (revised bstar prescription of https://arxiv.org/abs/hep-ph/0506225)
      if (opts.blimlha || opts.evolmode == 2)
	{
	  opts.bstar_pdf = true;
	  opts.blim_pdf = min(resconst::b0/pdf::qmin, opts.blim_pdf);
	}
      
      mc = LHAPDF::getThreshold(4);
      mb = LHAPDF::getThreshold(5);
      mt = LHAPDF::getThreshold(6);

      //Call PDFs once outside of multithreading (may be needed for LHAPDF initialisation?)
      double fPDF[13];
      pdf::xfxq(0.5,coupling::zmass,fPDF);

      opts.check_consistency();      
    }
  else
    {
      opts.totpdf = 1;
      pdferropts_.pdferr_ = false;
      pdferropts_.totpdf_ = 1;
      if (!photon)
	xfxqphot = lhaxfxqphot_empty;
    }

  // initialization of alphas
  //setalphas(opts.cmw); //--> Cannot initialise alphas here because scales::ren is not yet set

  //read g from the PDF
  setg();
  
  //take the cmass and b mass from the PDF
  //      cmass=dsqrt(mcsq)
  //      bmass=dsqrt(mbsq)
}

//run alphas
double pdf::alphas(double q, bool cmw)
{
  /**** obsolete running used in DYRes ****/
  //if (opts_.approxpdf_ == 1)
  //  {
  //    double asmz = extalphas(coupling::zmass);
  //    int nloop = 3;
  //    as = dyalphas_mcfm_(scales::ren,asmz,nloop);
  //  }
  /****************************************/

  //MSbar evolution
  double as;
  if (opts.alphaslha)
    as = extalphas(q);
  else
    as = rgktalphas(q);

  //Conversion to CMW
  if (cmw)
    switch (opts.order)
      {
      case 0: as = as; break;
      case 1: as = as*(1 + as/pi*resconst::K1 ); break;
      case 2: as = as*(1 + as/pi*resconst::K1 + pow(as/pi,2)*resconst::K2 ); break;
      case 3: as = as*(1 + as/pi*resconst::K1 + pow(as/pi,2)*resconst::K2 + pow(as/pi,3)*resconst::K3 ); break;
      case 4: as = as*(1 + as/pi*resconst::K1 + pow(as/pi,2)*resconst::K2 + pow(as/pi,3)*resconst::K3 + pow(as/pi,4)*resconst::K4 ); break;
      }

  //cout << "alphas(q) at q = " << q << "  " << as << endl;
  //cout << resconst::K1 << endl;
  
  return as;
}

//set value of alphas at the renormalization scale in the fortran common blocks
void pdf::setalphas(bool cmw)
{
  //run alphas
  double as = alphas(scales::ren, cmw);

  //Set alphas and strong coupling in the fortran common blocks

  //MCFM
  qcdcouple_.as_ = as;
  qcdcouple_.ason2pi_ = as/(2*M_PI);
  qcdcouple_.ason4pi_ = as/(4*M_PI);
  qcdcouple_.gsq_= 4*M_PI*as;

  //vjet analytical calculation
  asnew_.as_ = as/M_PI;
  asp_.asp_ = as;
}

//set the value of the g-parameter of the non perturbative form factor
void pdf::setg()
{
  if (!opts.externalpdf)
    {
      LHAPDF::PDFInfo info(opts.LHAPDFset, opts.LHAPDFmember);
      g1 = info.get_entry_as<double>("g", -1);
      g1 = info.get_entry_as<double>("g1", -1);
      g2 = info.get_entry_as<double>("g2", -1);
      e  = info.get_entry_as<double>("e", -1);
      g0  = info.get_entry_as<double>("g0", -1);
      q  = info.get_entry_as<double>("q", -1);
      blim  = info.get_entry_as<double>("blim", -1);
      lambda  = info.get_entry_as<double>("lambda", -1);
    }
  if (g1 >= 0)
    {
      opts.g1 = g1;
      g_param_.g_param_ = g1;
      np_.g_ = g1;
    }
  if (g2 >= 0)
    opts.g2 = g2;
  if (e >= 0)
    opts.e = e;
  if (g0 >= 0)
    opts.g0 = g0;
  if (q > -1)
    opts.q = q;
  if (blim >= 0)
    {
      opts.blim = blim;
      opts.blim_pdf     = blim;
      opts.blim_sudakov = blim;
      opts.blim_expc    = blim;
    }
  if (lambda > -1)
    opts.lambda = lambda;
  
  if (g1 >= 0 || g2 >= 0)
    cout << "g form factor: input from PDF member: g1 = " << g1 << "; g2 = " << g2 << endl;
  if (e >= 0)
    cout << "e form factor: input from PDF member: e = " << e << endl;
  if (g0 >= 0)
    cout << "g0 form factor: input from PDF member: g0 = " << g0 << endl;
  if (q > -1)
    cout << "q form factor: input from PDF member: q = " << q << endl;
  if (blim >= 0)
    cout << "blim: input from PDF member: blim = " << blim << endl;
  if (lambda > -1)
    cout << "lambda form factor: input from PDF member: lambda = " << lambda << endl;
}


void dysetpdf_(int& member)
{
  if (opts.externalpdf) return;
  if (opts.totpdf == 1) return;
  
  if (member == 0)
    {
//      if (opts.PDFerrors && opts.totpdf > 1)
//	LHAPDF::initPDF(0);
//      else
	LHAPDF::initPDF(opts.LHAPDFmember);
    }
  else
    LHAPDF::initPDF(member);
  
  pdf::setalphas(opts.cmw);
  //  setg(); //set g separately when setting a different PDF in the resummed part
}


void setmellinpdf_(int &member){
    // if member is still same than dont do anything
    //if (lastMember==member) return;
    //if (v_mellinpdf.size()<member) v_mellinpdf.
    // test current flag
    //if (v_mellinpdf[lastMember].isInitialized)
    // weights
    // moms
    // set init flag
    //lastMember=member;
}

void fdist_(int& ih, double& x, double& xmu, double fx[2*MAXNF+1])
{
  //set to zero if x out of range
  //check for nans
  if (isnan_ofast(x))
    {
      cout << "Bjorken-x is nan in fdist_" << endl;
      for (int i = -MAXNF; i <= MAXNF; i++)
	fx[MAXNF+i]=0.;
      return;
    }
  if (x > 1. || x <= 0.)
    {
      for (int i = -MAXNF; i <= MAXNF; i++)
	fx[MAXNF+i]=0.;
      return;
    }

  //Old LHAPDF interface
  //double fPDF[13];
  //LHAPDF::xfx(x,xmu,fPDF);

  //New LHAPDF interface
  //vector <double> fPDF; fPDF.resize(13);
  //pdf::lhapdf->xfxQ(x,xmu,fPDF);

  //Allows external PDF
  double fPDF[13];
  pdf::xfxq(x,xmu,fPDF);

  //cout << "fdist " << setprecision(16) << fPDF[6] << endl;
  //vector<int> pids = pdf::lhapdf->flavors();
  //for (int i = 0; i < pids.size(); i++)
  //cout << pids[i] << endl;

  //cout << " fdist x:" << x << " q: " << xmu << " - ";
  //for (int i = -MAXNF; i <= MAXNF; i++)
  //  cout << fPDF[6+i]/x << "  ";
  //cout << endl;
  
  //xmu *= xmu;
  //fPDF[6-5] = pdf::lhapdf->xfxQ(-5,x,xmu);
  //fPDF[6-4] = pdf::lhapdf->xfxQ(-4,x,xmu);
  //fPDF[6-3] = pdf::lhapdf->xfxQ(-3,x,xmu);
  //fPDF[6-2] = pdf::lhapdf->xfxQ(-2,x,xmu);
  //fPDF[6-1] = pdf::lhapdf->xfxQ(-1,x,xmu);
  //fPDF[6+0] = pdf::lhapdf->xfxQ(21,x,xmu);
  //fPDF[6+1] = pdf::lhapdf->xfxQ(1 ,x,xmu);
  //fPDF[6+2] = pdf::lhapdf->xfxQ(2 ,x,xmu);
  //fPDF[6+3] = pdf::lhapdf->xfxQ(3 ,x,xmu);
  //fPDF[6+4] = pdf::lhapdf->xfxQ(4 ,x,xmu);
  //fPDF[6+5] = pdf::lhapdf->xfxQ(5 ,x,xmu);

  //  for (int i = -MAXNF; i <= MAXNF; i++)
  //    cout << fPDF[6+i]/x << "  ";
  //  cout << endl;
  

  if (ih == 1) //proton
    for (int i = -MAXNF; i <= MAXNF; i++)
      fx[MAXNF+i]=fPDF[6+i]/x;
  else if (ih == -1) //antiproton
    for (int i = -MAXNF; i <= MAXNF; i++)
      fx[MAXNF+i]=fPDF[6-i]/x;

  //switch off flavours
  //fx[MAXNF-5]=0.; //bbar
  //fx[MAXNF-4]=0.; //cbar
  //fx[MAXNF-3]=0.; //sbar
  //fx[MAXNF-2]=0.;
  //fx[MAXNF-1]=0.;
  //fx[MAXNF+0]=0.; //gluon
  //fx[MAXNF+1]=0.;
  //fx[MAXNF+2]=0.;
  //fx[MAXNF+3]=0.; //s
  //fx[MAXNF+4]=0.; //c
  //fx[MAXNF+5]=0.; //b

  //impose positivity
  //fx[MAXNF-5]=max(0.,fx[MAXNF-5]);
  //fx[MAXNF-4]=max(0.,fx[MAXNF-4]);
  //fx[MAXNF-3]=max(0.,fx[MAXNF-3]);
  //fx[MAXNF-2]=max(0.,fx[MAXNF-2]);
  //fx[MAXNF-1]=max(0.,fx[MAXNF-1]);
  //fx[MAXNF+0]=max(0.,fx[MAXNF+0]);
  //fx[MAXNF+1]=max(0.,fx[MAXNF+1]);
  //fx[MAXNF+2]=max(0.,fx[MAXNF+2]);
  //fx[MAXNF+3]=max(0.,fx[MAXNF+3]);
  //fx[MAXNF+4]=max(0.,fx[MAXNF+4]);
  //fx[MAXNF+5]=max(0.,fx[MAXNF+5]);

  //make u = d and ubar = dbar
  //fx[MAXNF-1]=fx[MAXNF-2];
  //fx[MAXNF+1]=fx[MAXNF+2];

  /*
  if (x < 90./13000.*exp(-1.))
    {
      fx[MAXNF-5]=0.;
      fx[MAXNF-4]=0.;
      fx[MAXNF-3]=0.;
      fx[MAXNF-2]=0.;
      fx[MAXNF-1]=0.;
      fx[MAXNF+0]=0.;
      fx[MAXNF+1]=0.;
      fx[MAXNF+2]=0.;
      fx[MAXNF+3]=0.;
      fx[MAXNF+4]=0.;
      fx[MAXNF+5]=0.;
    }
  */

  //flat PDFs
  //fx[MAXNF-5]=1.; //bbar
  //fx[MAXNF-4]=1.; //cbar
  //fx[MAXNF-3]=1.; //sbar
  //fx[MAXNF-2]=1.;
  //fx[MAXNF-1]=1.;
  //fx[MAXNF+0]=1.; //gluon
  //fx[MAXNF+1]=1.;
  //fx[MAXNF+2]=1.;
  //fx[MAXNF+3]=1.; //s
  //fx[MAXNF+4]=1.; //c
  //fx[MAXNF+5]=1.; //b
}

// --> instead of overloading the fdist function, add a new function which retrieves only the photon PDF
void fdistphot_(int& beam, double& x, double& xmu, double &fotfx)
{
  pdf::fdistphot(beam, x, xmu, fotfx);
}

void pdf::fdistphot(int beam, double x, double mu, double& fotfx)
{
  pdf::xfxqphot(x,mu,fotfx);
  if (opts.nuclearpdf)
    {
      double a = 0.;
      if (beam == 1)
	a = opts.A1;
      if (beam == 2)
	a = opts.A2;
      fotfx *= a;
    }
}

void pdf_fdist_(int& beam, double& x, double& mu, double fx[2*MAXNF+1])
{
  pdf::fdist(beam, x, mu, fx);
}

//nuclear PDFs
void pdf::fdist(int beam, double x, double mu, double fx[2*MAXNF+1])
{
  //set to zero if x out of range
  if (x > 1. || x <= 0.)
    {
      for (int i = -MAXNF; i <= MAXNF; i++)
	fx[MAXNF+i]=0.;
      return;
    }

  //Allows external PDF
  double fPDF[14] = {0.};
  if (beam == 2 && secondbeam)
    lhaxfxq_beam2(x,mu,fPDF);
  else
    xfxq(x,mu,fPDF);

  if (!opts.nuclearpdf)
    {
      //proton, antiproton or pions
      int ih;
      if (beam == 1)
	ih = opts.ih1;
      if (beam == 2)
	ih = opts.ih2;

      if (ih == 1 || ih == -211 || fabs(ih) > 2212) //proton or negative pion (xFitter or Jam)
	for (int i = -MAXNF; i <= MAXNF; i++)
	  fx[MAXNF+i]=fPDF[6+i]/x;
      else if (ih == -1 || ih == 211) //antiproton
	for (int i = -MAXNF; i <= MAXNF; i++)
	  fx[MAXNF+i]=fPDF[6-i]/x;
      else if (ih == -221)// pi- a la GRV
	{
	  for (int i = -MAXNF; i <= MAXNF; i++)
	    fx[MAXNF+i]=fPDF[6+i]/x;
	  fx[MAXNF-2]=fx[MAXNF+2];//ubar is becoming valence
	  fx[MAXNF+2]=fx[MAXNF-1];//u is becoming sea contribution
	  //do I have to add sea compontents to valence?
	}
      else if (ih == 221)//pi+ a la GRV
	{
	  for (int i = -MAXNF; i <= MAXNF; i++)
	    fx[MAXNF+i]=fPDF[6+i]/x;
	  fx[MAXNF-1]=fx[MAXNF+1];//dbar is becoming valence
	  fx[MAXNF+1]=fx[MAXNF-2];//d is becoming sea contribution
	  //do I have to add sea components to valence?
	}
    }      
  else
    {
      //number of protons
      double p;
      if (beam == 1)
	p = opts.Z1;
      if (beam == 2)
	p = opts.Z2;

      //number of neutrons
      double n = 0;
      if (beam == 1)
	n = opts.A1-opts.Z1;
      if (beam == 2)
	n = opts.A2-opts.Z2;
  
      double fxp[2*MAXNF+1];
      double fxn[2*MAXNF+1];
      for (int i = -MAXNF; i <= MAXNF; i++)
	{
	  fxp[MAXNF+i]=fPDF[6+i]/x;
	  fxn[MAXNF+i]=fPDF[6+i]/x;
	}
      fxn[parton::U]  = fxp[parton::D];
      fxn[parton::D]  = fxp[parton::U];
      fxn[parton::Ub] = fxp[parton::Db];
      fxn[parton::Db] = fxp[parton::Ub];

      for (int i = -MAXNF; i <= MAXNF; i++)
	fx[MAXNF+i]=p*fxp[MAXNF+i]+n*fxn[MAXNF+i];
      //      for (int i = -MAXNF; i <= MAXNF; i++)
      //	cout << fx[MAXNF+i] << endl;
    }
}
